package frc.robot.subsystems;

import java.util.function.DoubleSupplier;

import com.revrobotics.CANSparkMax;
import com.revrobotics.RelativeEncoder;
import com.revrobotics.CANSparkLowLevel.MotorType;

import edu.wpi.first.math.controller.ArmFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.math.controller.SimpleMotorFeedforward;
import edu.wpi.first.wpilibj.Encoder;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.ShooterConstants;

public class Shooter extends SubsystemBase{
    private CANSparkMax topShooter;
    private CANSparkMax bottomShooter;

    private RelativeEncoder bottomEncoder;
    private RelativeEncoder topEncoder;

    private CANSparkMax shooterPivot;

    private Encoder pivotEncoder;

    private PIDController topShooterPID;
    private SimpleMotorFeedforward topShooterFF;

    private PIDController bottomShooterPID;
    private SimpleMotorFeedforward bottomShooterFF;

    private PIDController shooterPivotPID;
    private ArmFeedforward shooterPivotFF;

    public Shooter(){
        topShooter = new CANSparkMax(ShooterConstants.kTopShooterID, MotorType.kBrushless);
        bottomShooter = new CANSparkMax(ShooterConstants.kBottomShooterID, MotorType.kBrushless);

        topEncoder = topShooter.getEncoder();
        bottomEncoder = bottomShooter.getEncoder();

        shooterPivot = new CANSparkMax(ShooterConstants.kShooterPivotID, MotorType.kBrushless);

        pivotEncoder = new Encoder(0, 1);
        pivotEncoder.setDistancePerPulse(ShooterConstants.kPivotConversion);

        //TODO These can probably devolve into BangBang controllers and let the feed forwards maintain speed
        topShooterPID = new PIDController(
            ShooterConstants.kShooterP, 
            ShooterConstants.kShooterI, 
            ShooterConstants.kShooterD
        );
        bottomShooterPID = new PIDController(
            ShooterConstants.kShooterP, 
            ShooterConstants.kShooterI, 
            ShooterConstants.kShooterD
        );

        topShooterFF = new SimpleMotorFeedforward(ShooterConstants.kSShooterFF, ShooterConstants.kVShooterFF);
        topShooterFF = new SimpleMotorFeedforward(ShooterConstants.kSShooterFF, ShooterConstants.kVShooterFF);

        shooterPivotPID = new PIDController(
            ShooterConstants.kShooterPivotP,
            ShooterConstants.kShooterPivotI,
            ShooterConstants.kShooterPivotD
        );

        shooterPivotFF = new ArmFeedforward(
            ShooterConstants.kSShooterPivotFF,
            ShooterConstants.kGShooterPivotFF, 
            ShooterConstants.kVShooterPivotFF
        );
    }

    public Command angleSpeedsSetpoints(double setpointAngle, double topRPM, double bottomRPM){
        return run(()-> {
            shooterPivot.setVoltage(
                shooterPivotPID.calculate(pivotEncoder.getDistance(), setpointAngle) + 
                shooterPivotFF.calculate(setpointAngle, 0.0));

            topShooter.setVoltage(topShooterPID.calculate(topEncoder.getVelocity(), topRPM)+topShooterFF.calculate(topRPM));

            bottomShooter.setVoltage(bottomShooterPID.calculate(bottomEncoder.getVelocity(), bottomRPM)+bottomShooterFF.calculate(bottomRPM));
        });
    }

    public DoubleSupplier getShooterAngle(){
        return pivotEncoder.getDistance();
    }
}
